from typing import Union


class ApiResponse:
    def __init__(self, data: dict) -> None:
        try:
            self.data = data
        except AttributeError:
            self.data = None

    
    @property
    def message(self) -> Union[str, None]:
        """Returns the message of the API response."""
        return self.data.get("api:message")
    
    @property
    def status_code(self) -> Union[int, None]:
        """Returns the status code of the API response."""
        return self.data.get("api:statuscode")
    
    @property
    def duration(self) -> Union[str, None]:
        """Returns the duration of the API response."""
        return self.data.get("api:duration")
    
    @property
    def timestamp(self) -> Union[str, None]:
        """Returns the timestamp of the API response."""
        return self.data.get("api:timestamp")
    
    @property
    def media_value(self) -> Union[str, None]:
        """Returns the media value of the API response."""
        return self.data.get("mediaValue") or self.data.get("result", {}).get("mediaValue")
    
    @property
    def mediaValue(self) -> Union[str, None]:
        #NOTE: This will be removed in the future.
        """Returns the media value of the API response."""
        return self.media_value
    
    def json(self) -> Union[dict, str]:
        """Returns the JSON data of the API response."""
        return self.data
    
    def __repr__(self) -> str:
        """Returns the representation of the API response."""
        return f"<ApiResponse status_code={self.status_code} message={self.message} media_value={self.media_value}>"


from uuid import uuid4
from json import loads, dumps
from colorama import Fore, Style
from typing import Optional, Union, Tuple, Callable

from .generate import Generator
from ..entities.handlers import orjson_exists
from requests import Session as Http, Response as HttpResponse

from ..entities import (
    Forbidden,
    BadGateway,
    APIException,
    ServiceUnavailable
    )

from requests.exceptions import (
    ConnectionError,
    ReadTimeout,
    SSLError,
    ProxyError,
    ConnectTimeout
    )

if orjson_exists():
    from orjson import (
        loads as orjson_loads,
        dumps as orjson_dumps
        )

class RequestHandler:
    """
    `RequestHandler` - A class that handles all requests

    `**Parameters**``
    - `bot` - The main bot class.
    - `generator` - The generator class.    
    - `proxy` - The proxy to use for requests.

    """
    def __init__(
        self,
        bot,
        generator: Generator,
        KEY: Optional[str] = None,
        proxy: Optional[str] = None
        ) -> None:
        self.bot             = bot
        self.generate        = generator
        self.api_url:        str = "http://service.aminoapps.com/api/v1"
        self.http_handler:   Http = Http()
        self.sid:            Optional[str] = None
        self.device:         Optional[str] = None
        self.userId:         Optional[str] = None
        self.email:          Optional[str] = None
        self.password:       Optional[str] = None
        self.__key__:        Optional[str] = KEY
        self.orjson:         bool = orjson_exists()

        self.proxy = {
            "http": proxy,
            "https": proxy
            } if proxy is not None else None

        self.response_map = {
            403: Forbidden,
            502: BadGateway,
            503: ServiceUnavailable
            }

    def service_url(self, url: str) -> str:
        """
        `service_url` - Appends the endpoint to the service url
        
        `**Parameters**``
        - `url` - The endpoint to append to the service url.
        
        `**Returns**``
        - `str` - The service url.
        
        """
        return f"{self.api_url}{url}" if url.startswith("/") else url
    
    def service_headers(self) -> dict:
        """Returns the service headers"""
        return {
            "NDCLANG": "en",
            "ACCEPT-LANGUAGE": "en-US",
            "User-Agent": "Dalvik/2.1.0 (Linux; U; Android 12; com.narvii.amino.master/3.5.35071)",
            "HOST": "service.aminoapps.com",
            "CONNECTION": "Keep-Alive",
            "ACCEPT-ENCODING": "gzip, deflate, br"
            }
    
    def fetch_request(self, method: str) -> Callable:
        """
        `fetch_request` - Returns the request method
        
        `**Parameters**``
        - `method` - The request method to return.
        
        `**Returns**``
        - `Callable` - The request method.
        
        """
        request_methods = {
            "GET": self.http_handler.get,
            "POST": self.http_handler.post,
            "DELETE": self.http_handler.delete,
            }
        return request_methods[method]
    
    def send_request(
            self,
            method: str,
            url: str,
            data: Union[dict, bytes, None],
            headers: dict,
            content_type: Optional[str]
        ) -> Tuple[int, str]:
        """
        `send_request` - Sends a request
        
        `**Parameters**``
        - `method` - The request method to use.
        - `url` - The url to send the request to.
        - `data` - The data to send with the request.
        - `headers` - The headers to send with the request.
        - `content_type` - The content type of the data.
        
        `**Returns**``
        - `Tuple[int, str]` - The status code and response from the request.
        
        """
        try:
            response: HttpResponse = self.fetch_request(method)(
                url, data=data, headers=headers, proxies=self.proxy
            )
            return response.status_code, response.text
        except (
            ConnectionError,
            ReadTimeout,
            SSLError,
            ProxyError,
            ConnectTimeout,
        ) as e:
            self.bot._log(f"Failed to send request: {e}")
            return self.handler(method, url, data, content_type)

    def handler(
        self,
        method: str,
        url: str,
        data: Union[dict, bytes, None] = None,
        content_type: Optional[str] = None,
        is_login_required: bool = True
    ) -> dict:
        """
        `handler` - Handles all requests
        
        `**Parameters**``
        - `method` - The request method to use.
        - `url` - The url to send the request to.
        - `data` - The data to send with the request.
        - `content_type` - The content type of the data.
        - `is_login_required` - Whether or not the request requires a login.
        
        `**Returns**``
        - `dict` - The response from the request.
        
        """
    
        url = self.service_url(url)

        if data is not None and isinstance(data, dict):
            data = {**data, "uid": self.userId}

        url, headers, binary_data = self.service_handler(url, data, content_type)

        if all([method=="POST", data is None]):
            headers["CONTENT-TYPE"] = "application/octet-stream"

        if not is_login_required:
            headers.pop("NDCAUTH")
            headers.pop("AUID")

        status_code, content = self.send_request(
            method, url, binary_data, headers, content_type
        )

        self.print_response(method=method, url=url, status_code=status_code)

        return self.handle_response(status_code=status_code, response=content, url=url)


    def service_handler(
        self,
        url: str,
        data: Union[dict, bytes, None] = None,
        content_type: Optional[str] = None
    ) -> Tuple[str, dict, Union[dict, bytes, None]]:
        """
        `service_handler` - Signs the request and returns the service url, headers and data

        `**Parameters**``
        - `url` - The url to send the request to.
        - `data` - The data to send with the request.
        - `content_type` - The content type of the data.
        
        `**Returns**``
        - `Tuple[str, dict, Union[dict, bytes, None]]` - The service url, headers and data.

        """
        
        if data or content_type:
            headers, data = self.fetch_signature(data, headers, content_type)

        return url, headers, self.ensure_utf8(data)

    def ensure_utf8(self, data: Union[dict, bytes, None]) -> Union[dict, bytes, None]:
        """
        `ensure_utf8` - Ensures the data is utf-8 encoded
        
        `**Parameters**``
        - `data` - The data to encode.
        
        `**Returns**``
        - `Union[dict, bytes, None]` - The encoded data.
        
        """

        if data is None: return data

        def handle_dict(data: dict):
            return {key: self.ensure_utf8(value) for key, value in data.items()}

        def handle_str(data: str):
            return data.encode("utf-8")

        handlers = {
            dict: handle_dict,
            str: handle_str
        }

        return handlers.get(type(data), lambda x: x)(data)

    def fetch_signature(
        self,
        data: Union[dict, bytes, None],
        headers: dict,
        content_type: str = None
    ) -> Tuple[dict, Union[dict, bytes, None]]:
        """
        `fetch_signature` - Fetches the signature and returns the data and updated headers

        `**Parameters**``
        - `data` - The data to send with the request.
        - `headers` - The headers to send with the request.
        - `content_type` - The content type of the data.

        `**Returns**``
        - `Tuple[dict, Union[dict, bytes, None]]` - The headers and data.

        """

        if not isinstance(data, bytes):
            data = orjson_dumps(data).decode("utf-8") if self.orjson else dumps(data)
        
        response = self.generate.NdcMessageSignature(data, self.userId)
        if response is None:
            SID, NDCDEVICEID, NDC_MESSAGE_SIGNATURE = None, self.device or self.generate.device_id(), None
        else:
            SID, NDCDEVICEID, NDC_MESSAGE_SIGNATURE = response["0"], response["1"], response["2"]
            SID = "sid={}".format(SID)
        headers.update({
            "CONTENT-LENGTH": str(len(data)),
            "CONTENT-TYPE": content_type or "application/json; charset=utf-8",
            "NDC-MSG-SIG": (
                self.generate.signature(data)
            ),
            "NDC-MESSAGE-SIGNATURE": NDC_MESSAGE_SIGNATURE,
            "NDCAUTH": SID,
            "AUID": self.userId or str(uuid4()),
            "NDCDEVICEID": NDCDEVICEID
        })
        return headers, data
        
    def handle_response(self, status_code: int, response: str, url: str) -> dict:
        """
        `handle_response` - Handles the response and returns the response as a dict
        
        `**Parameters**``
        - `status_code` - The status code of the response.
        - `response` - The response to handle.
        
        `**Returns**``
        - `dict` - The response as a dict.
        
        """
        if status_code in self.response_map:
            raise self.response_map[status_code]
        
        response = loads(response)
        
        if status_code != 200:
            raise APIException(response)

        return response

    def print_response(self, method: str, url: str, status_code: int):
        """
        `print_response` - Prints the response if debug is enabled

        `**Parameters**``
        - `method` - The request method used.
        - `url` - The url the request was sent to.
        - `status_code` - The status code of the response.
        - `response` - The response to print.

        """
        if self.bot.debug:
            color = Fore.RED if status_code != 200 else {
                "GET": Fore.BLUE,
                "POST": Fore.GREEN,
                "DELETE": Fore.MAGENTA,
                "LITE": Fore.YELLOW
                }.get(method, Fore.RED)
            print(f"{color}{Style.BRIGHT}{method}{Style.RESET_ALL} - {url}")


from typing import List


class ExtData:
    def __init__(self, data: dict) -> None:
        try:
            self.data = data
        except AttributeError:
            self.data = None

    def _check_ext_data(F):
        def wrapper(*args, **kwargs):
            return None if args[0].data is None else F(*args, **kwargs)
        return wrapper

    @property
    @_check_ext_data
    def value(self) -> dict:
        """Returns the value of the ext data"""
        return self.data.get("value")   

    @property
    @_check_ext_data
    def note(self) -> int:
        """Returns the note the moderator/admin left on the request"""
        return self.data.get("note")
    
    def json(self) -> dict:
        """Returns the raw json data"""
        return self.data
    

class ExtDataList:
    def __init__(self, data: list) -> None:
        try:
            self.data = data
        except AttributeError:
            self.data = None

    def _check_ext_data_list(F):
        def wrapper(*args, **kwargs):
            return None if args[0].data is None else F(*args, **kwargs)
        return wrapper

    @property
    @_check_ext_data_list
    def value(self) -> List[dict]:
        """Returns a list of values of the ext data"""
        return [data.get("value") if data is not None else None for data in self.data]

    @property
    @_check_ext_data_list
    def note(self) -> List[str]:
        """Returns a list of notes the moderator/admin left on the request"""
        return [str(note.get("note")) if note is not None else None for note in self.data]
    
    def json(self) -> List[dict]:
        """Returns the raw json data"""
        return self.data


class AdminLog:
    def __init__(self, data: dict) -> None:
        try:
            self.data = data
        except AttributeError:
            self.data = None

    def _check_admin_log(F):
        def wrapper(*args, **kwargs):
            return None if args[0].data is None else F(*args, **kwargs)
        return wrapper

    @property
    @_check_admin_log
    def operation_name(self) -> str:
        """Returns the name of the operation"""
        return self.data.get("operationName")

    @property
    @_check_admin_log
    def comId(self) -> int:
        """Returns the comId of the community the operation was performed on"""
        return self.data.get("comId")

    @property
    @_check_admin_log
    def refer_ticket_id(self) -> int:
        """Returns the ticket id of the operation"""
        return self.data.get("referTicketId")

    @property
    @_check_admin_log
    def object_url(self) -> str:
        """Returns the url of the object the operation was performed on"""
        return self.data.get("objectUrl")

    @property
    @_check_admin_log
    def created_time(self) -> str:
        """Returns the time the operation was performed"""
        return self.data.get("createdTime")

    @property
    @_check_admin_log
    def ext_data(self) -> ExtData:
        """Returns the ext data of the operation"""
        return ExtData(self.data.get("extData"))

    @property
    @_check_admin_log
    def operation_level(self) -> int:
        """Returns the level of the operation"""
        return self.data.get("operationLevel")

    @property
    @_check_admin_log
    def operation_id(self) -> int:
        """Returns the id of the operation"""
        return self.data.get("operation")

    @property
    @_check_admin_log
    def object_type(self) -> int:
        """Returns the type of the object the operation was performed on"""
        return self.data.get("objectType")

    @property
    @_check_admin_log
    def operation_details(self) -> str:
        """Returns the details of the operation"""
        return self.data.get("operationDetail")

    @property
    @_check_admin_log
    def log_id(self) -> int:
        """Returns the log id of the operation"""
        return self.data.get("logId")

    @property
    @_check_admin_log
    def moderation_level(self) -> int:
        """Returns the moderation level of the operation"""
        return self.data.get("moderationLevel")

    @property
    @_check_admin_log
    def objectId(self) -> int:
        """Returns the object id of the operation"""
        return self.data.get("objectId")

    @property
    @_check_admin_log
    def moderator(self) -> dict:
        """Returns the moderator of the operation"""
        return self.data.get("author")

    @property
    @_check_admin_log
    def moderator_username(self) -> str:
        """Returns the username of the moderator of the operation"""
        return None if self.moderator is None else self.moderator.get("nickname")

    @property
    @_check_admin_log
    def moderator_uid(self) -> str:
        """Returns the uid of the moderator of the operation"""
        return None if self.moderator is None else self.moderator.get("uid")

    @property
    @_check_admin_log
    def moderator_icon(self) -> str:
        """Returns the icon of the moderator of the operation"""
        return None if self.moderator is None else self.moderator.get("icon")

    def json(self) -> dict:
        """Returns the raw json data"""
        return self.data
    

class AdminLogList:
    def __init__(self, data: dict) -> None:
        try:
            self.data = data
        except AttributeError:
            self.data = None

    def parser(self) -> List[AdminLog]:
        """Returns a list of AdminLog objects"""
        return [AdminLog(i) for i in self.data.get("adminLogList")]
    
    @property
    def paging(self) -> dict:
        """Returns the paging data"""
        return self.data.get("paging")
    
    @property
    def next_page_token(self) -> int:
        """Returns the next page token"""
        return self.paging.get("nextPageToken")
    
    @property
    def prev_page_token(self) -> int:
        """Returns the previous page token"""
        return self.paging.get("prevPageToken")
    
    @property
    def operation_name(self) -> List[str]:
        """Returns a list of operation names"""
        return [i.operation_name for i in self.parser()]

    @property
    def comId(self) -> List[int]:
        """Returns a list of comIds"""
        return [i.comId for i in self.parser()]

    @property
    def refer_ticket_id(self) -> List[int]:
        """Returns a list of ticket ids"""
        return [i.refer_ticket_id for i in self.parser()]

    @property
    def object_url(self) -> List[str]:
        """Returns a list of object urls"""
        return [i.object_url for i in self.parser()]

    @property
    def created_time(self) -> List[str]:
        """Returns a list of created times"""
        return [i.created_time for i in self.parser()]
    
    @property
    def ext_data(self) -> ExtDataList:
        """Returns a list of ExtData objects"""
        return ExtDataList([i.ext_data.json() for i in self.parser()])

    @property
    def operation_level(self) -> List[int]:
        """Returns a list of operation levels"""
        return [i.operation_level for i in self.parser()]

    @property
    def operation_id(self) -> List[int]:
        """Returns a list of operation ids"""
        return [i.operation_id for i in self.parser()]
    
    @property
    def object_type(self) -> List[int]:
        """Returns a list of object types"""
        return [i.object_type for i in self.parser()]

    @property
    def operation_details(self) -> List[str]:
        """Returns a list of operation details"""
        return [i.operation_details for i in self.parser()]

    @property
    def log_id(self) -> List[int]:
        """Returns a list of log ids"""
        return [i.log_id for i in self.parser()]

    @property
    def moderation_level(self) -> List[int]:
        """Returns a list of moderation levels"""
        return [i.moderation_level for i in self.parser()]

    @property
    def objectId(self) -> List[int]:
        """Returns a list of object ids"""
        return [i.objectId for i in self.parser()]

    @property
    def moderator(self) -> List[dict]:
        """Returns a list of moderators"""
        return [i.moderator for i in self.parser()]

    @property
    def moderator_username(self) -> List[str]:
        """Returns a list of moderator usernames"""
        return [i.moderator_username for i in self.parser()]
    
    @property
    def moderator_uid(self) -> List[str]:
        """Returns a list of moderator uids"""
        return [i.moderator_uid for i in self.parser()]

    @property
    def moderator_icon(self) -> str:
        """Returns a list of moderator icons"""
        return [i.moderator_icon for i in self.parser()]

    def json(self) -> dict:
        """Returns the raw json data"""
        return self.data


from hmac import new
from hashlib import sha1
from base64 import b64encode
from secrets import token_hex
from typing import Union
import requests
from time import sleep

class Generator:
    def __init__(
        self,
        prefix:        Union[str, int],
        device_key:    str,
        signature_key: str,
        KEY: str
        ) -> None:
        self.PREFIX = bytes.fromhex(str(prefix))
        self.DEVICE_KEY = bytes.fromhex(device_key)
        self.SIGNATURE_KEY = bytes.fromhex(signature_key)
        self.KEY = KEY

    def device_id(self) -> str:
        """
        `generate_device_id` Generates a device ID based on a specific string.

        `**Returns**`
        - `str` - Returns a device ID as a string.
        """
        encoded_data = sha1(str(token_hex(20)).encode('utf-8')).hexdigest()

        digest = new(
            self.DEVICE_KEY,
            self.PREFIX + bytes.fromhex(encoded_data),
            sha1).hexdigest()

        return f"{bytes.hex(self.PREFIX)}{encoded_data}{digest}".upper()

    def signature(self, data: str) -> str:
        """
        `signature` Generates a signature based on a specific string.
        
        `**Parameters**`
        - `data` - Data to generate a signature from
        `**Returns**`
        - `str` - Returns a signature as a string.
        """

        signature = [self.PREFIX[0]]  
        signature.extend(new(
            self.SIGNATURE_KEY,
            str(data).encode("utf-8"), sha1).digest())

        return b64encode(bytes(signature)).decode("utf-8")
    
    def update_device(self, device: str) -> str:
        """
        Update a device ID to new prefix.

        :param device: The device ID to update.
        :type device: str
        :return: The updated device ID as a string.
        :rtype: str
        """
        encoded_data = sha1(str(bytes.fromhex(device[2:42])).encode('utf-8')).hexdigest()

        digest = new(
            self.DEVICE_KEY,
            self.PREFIX + bytes.fromhex(encoded_data),
            sha1).hexdigest()

        return f"{bytes.hex(self.PREFIX)}{encoded_data}{digest}".upper()
    
    def NdcMessageSignature(self, data: str, auid: str) -> Union[str, None]:
        if auid is None:
            return None
        
        response = requests.post(
            url="https://app.friendify.ninja/api/v1/pymino",
            headers={
                "AUID": auid,
                "KEY": self.KEY
            },
            data=str(data).encode("utf-8")
        )
        if response.status_code == 200:
            return response.json()
        elif response.status_code == 503:
            print(response.text)
            sleep(30)
            return self.NdcMessageSignature(
                data=data,
                auid=auid
            )
        else:
            raise Exception(
                response.text
            )


from colorama import Fore, Style

class Menu:
    def __init__(self, console):
        """
        The Menu class handles the main menu of the application.

        :param console: An instance of the Console class for inter-component communication.
        :type console: Console
        """
        self.console = console
        self.menu_logo = """
██████╗ ██╗   ██╗███╗   ███╗██╗███╗   ██╗ ██████╗ 
██╔══██╗╚██╗ ██╔╝████╗ ████║██║████╗  ██║██╔═══██╗
██████╔╝ ╚████╔╝ ██╔████╔██║██║██╔██╗ ██║██║   ██║
██╔═══╝   ╚██╔╝  ██║╚██╔╝██║██║██║╚██╗██║██║   ██║
██║        ██║   ██║ ╚═╝ ██║██║██║ ╚████║╚██████╔╝
╚═╝        ╚═╝   ╚═╝     ╚═╝╚═╝╚═╝  ╚═══╝ ╚═════╝ 
"""
        self.menu_logo = Fore.MAGENTA + self.menu_logo + Style.RESET_ALL
        self.author = "CONSOLE v0.1                    by @forevercynical"
        self.author = Fore.LIGHTYELLOW_EX + self.author + Style.RESET_ALL
        self.menu_logo += self.author

    def display(self):
        """
        Displays the main menu options to the user and processes their input.
        """
        self.console.clear()
        self.console.print(self.menu_logo)
        self.console.print(self.welcome_screen())
        self.console.print("""
    1. Select Community
    2. Edit Profile
    3. Join Public Chat
    4. My Chats
    """)
        choice = self.console.input(">>> ")
        print()
        menu = {
            "1": self.console.community_console.select_community,
            "2": self.console.profile_console.edit_profile,
            "3": self.console.chat_console.join_public_chat,
            "4": self.console.chat_console.my_chats
        }
        if choice in menu:
            self.console.clear()
            menu[choice]()
        else:
            self.console.print("Invalid option. Please try again.")
            self.display()

    def welcome_screen(self):
        """
        Returns a string containing the welcome message for the user.

        :return: A string containing the welcome message for the user.
        :rtype: str
        """
        profile = self.console.bot.profile
        userId = profile.userId
        aminoId = profile.aminoId
        username = profile.nickname
        selected_community = self.console.bot.community_id

        return f"""
    Welcome, {username}({aminoId})!
    User ID: {userId}
    Selected Community: {selected_community}
    """



import signal
from random import randint
from typing import Optional
from threading import Thread
from contextlib import suppress
from urllib.parse import urlencode
from time import sleep as delay, time
from json import loads, dumps, JSONDecodeError

from .entities import *
from .context import EventHandler
from .dispatcher import MessageDispatcher

if orjson_exists():
    from orjson import (
        loads as orjson_loads, dumps as orjson_dumps
        )

try:
    from websocket import WebSocket, WebSocketApp
except ImportError as e:
    pipmain(["uninstall", "websocket", "-y"])
    pipmain(["install", "websocket-client==1.6.1"])
    raise WrongWebSocketPackage from e


class WSClient(EventHandler):
    """
    WSClient class that handles websocket events.

    This class extends `EventHandler` class, allowing the bot to use event handler features.

    Special Attributes:
    __slots__ : tuple
        A tuple containing a fixed set of attributes to optimize memory usage.

    Attributes:
    ws : WebSocketApp
        The websocket object.
    _communities : set
        A set containing the communities the bot is in.
    event_types : dict
        A dictionary containing the event types.
    notif_types : dict
        A dictionary containing the notification types.
    dispatcher : MessageDispatcher
        The message dispatcher object.
    channel : Optional[Channel] 
        The agora channel.
    orjson : bool
        Whether or not orjson is installed.

    """
    __slots__ = (
        "ws",
        "_communities",
        "event_types",
        "is_logging",
        "notif_types",
        "dispatcher",
        "channel",
        "orjson"
    )
    def __init__(self):
        self.ws:            WebSocketApp = None
        self._communities:  set = set()
        self.event_types:   dict =  EventTypes().events
        self.is_logging:    bool = bool(self.logger)
        self.notif_types:   dict =  NotifTypes().notifs
        self.dispatcher:    MessageDispatcher = MessageDispatcher()
        self.channel:       Optional[Channel] = None
        self.orjson:        bool = orjson_exists()

        self.dispatcher.register(10, self._handle_notification)
        self.dispatcher.register(201, self._handle_agora_channel)
        self.dispatcher.register(400, self._handle_user_online)
        self.dispatcher.register(1000, self._handle_message)

        signal.signal(signal.SIGINT, signal.SIG_DFL)
        EventHandler.__init__(self)

    def fetch_ws_url(self) -> str:
        return f"wss://ws{randint(1, 4)}.aminoapps.com"
    
    def _log(self, message: str) -> None:
        """
        Logs a message to debug.log

        :param message: The message to log.
        :type message: str
        :return: None

        """
        if self.is_logging:
            try:
                self.logger.debug(message)
            except Exception:
                self.is_logging = False

    def connect(self) -> None:
        """Connects to the websocket."""
        self.run_forever()
        return self.emit("ready")

    def run_forever(self) -> None:
        """Runs the websocket forever."""
        self._log("Initializing websocket.")
        ws_data = f"{self.generate.device_id()}|{int(time() * 1000)}"
        self.ws = WebSocketApp(
            url = f"{self.fetch_ws_url()}/?{urlencode({'signbody': ws_data})}",
            on_open=self.on_websocket_open,
            on_message=self.on_websocket_message,
            on_error=self.on_websocket_error,
            on_close=self.on_websocket_close,
            header={
            "NDCDEVICEID": self.generate.device_id(),
            "NDCAUTH": f"sid={self.sid}",
            "NDC-MSG-SIG": self.generate.signature(ws_data)
            })
        
        self.start_processes()
        return self._log("Websocket connected.")

    def start_processes(self) -> None:
        """Starts the websocket processes."""
        websocket_thread = Thread(target=self.ws.run_forever)
        websocket_thread.start()

        aalive_thread = Thread(target=run_alive_loop, args=(self,))
        aalive_thread.start()

    def on_websocket_error(self, ws: WebSocket, error: Exception) -> None:
        """Handles websocket errors."""
        with suppress(KeyError):
            self._events["error"](error)

        return self._log(f"Websocket error: {error}")

    def on_websocket_message(self, ws: WebSocket, message: str) -> None:
        """Receives websocket messages."""
        return Thread(target=self._handle_websocket_message, args=(message,)).start()

    def _handle_websocket_message(self, message: str) -> None:
        """Handles websocket messages."""        
        try:
            raw_message = orjson_loads(message) if self.orjson else loads(message)
        except JSONDecodeError:
            raw_message = loads(message)

        return self.dispatcher.handle(raw_message)

    def _handle_message(self, message: dict) -> None:
        """Sends the message to the event handler."""
        _message: Message = Message(message)

        if self.userId == _message.userId: return None
        None if any(
            [_message.ndcId is None, _message.ndcId == 0]
        ) else self._communities.add(_message.ndcId)

        key = self.event_types.get(f"{_message.type}:{_message.mediaType}")

        return self._handle_event(key, _message) if key else None

    def _handle_notification(self, message: dict) -> None:
        """Handles notifications."""
        notification: Notification = Notification(message)
        key = self.notif_types.get(notification.notification_type)
        
        return self._handle_event(key, notification) if key else None

    def _handle_agora_channel(self, message: dict) -> None:
        """Sets the agora channel."""
        self.channel: Channel = Channel(message)

    def _handle_user_online(self, message: dict) -> None:
        """Handles user online events."""
        return self._handle_event("user_online", OnlineMembers(message))

    def on_websocket_close(self, ws: WebSocket, close_status_code: int, close_msg: str) -> None:
        """Handles websocket close events."""
        if [close_status_code, close_msg] == [None, None]:
            self._log("Websocket closed unexpectedly.")
            return self.run_forever()

    def send_websocket_message(self, message: dict) -> None:
        """Sends a websocket message."""
        return self.ws.send(orjson_dumps(message).decode() if self.orjson else dumps(message))

    def stop_websocket(self) -> None:
        """Stops the websocket."""
        self._log("Websocket received stop signal.")
        return self.ws.close()

    def on_websocket_open(self, ws: WebSocket) -> None:
        """Handles websocket open events."""
        if all([self.community_id != None, "user_online" in self._events]):
            return self.send_websocket_message({
                "t": 300,
                "o": {
                    "ndcId": self.community_id,
                    "topic": f"ndtopic:x{self.community_id}:online-members",
                    "id": int(time() * 1000)
                }})

    def _last_active(self, last_activity_time: float) -> bool:
        """Returns True if the last activity was 5 minutes ago."""""
        return time() - last_activity_time >= 300

    def _last_message(self, last_message_time: float) -> bool:
        """Returns True if the last message was 30 seconds ago."""
        return time() - last_message_time >= 30

    def _send_message(self) -> None:
        self.send_websocket_message({
            "o":{
                "threadChannelUserInfoList": [],
                "id": randint(1, 100)},
                "t": 116
                })

    def _activity_status(self) -> None:
        """Sets the user's activity status to online."""
        for comId in self._communities:

            if self.online_status:
                try:
                    self.community.send_active(comId=comId,
                    timers=[{"start": int(time()), "end": int(time()) + 300}]
                    )
                except Exception:
                    self.online_status = False

            delay(randint(5, 10))